import ast
import copy

def dump(tree):
    print(ast.dump(tree))

def dump_code(node):
    print(ast.unparse(node))

def new_ast_const(val):
    return ast.Constant(value=val)

def new_ast_for(target, range_node, body=None):
    realbody = body if body is not None else []
    loop = ast.For(target=target, iter=range_node, body=realbody, lineno=None,
        orelse=[], type_ignores=[])
    return loop

def new_ast_perfect_for(targets, range_nodes, body=None):
    parent = root = ast.Module(body=[])
    for tar, rg in zip(targets, range_nodes):
        loop = new_ast_for(tar, rg)
        parent.body.append(loop)
        parent = loop
    
    if body is not None:
        parent.body.extend(body)
    return root.body[0]

def new_ast_call(name_node, args, keywords=None):
    kws = []
    if keywords:
        for k, v in keywords.items():
            kws.append(ast.keyword(arg=k, value=v))
    if type(args) != list:
        args = [args]
    node = ast.Call(func=name_node, args=args, keywords=kws)
    return node

def new_ast_range(stop, start=new_ast_const(0), step=new_ast_const(1)):
    node = ast.Call(func=new_ast_name('range'), args=[start, stop, step], keywords=[])
    return node

def new_ast_node_from_str(s, ctx=ast.Load(), inline=True):
    expr = ast.parse(s).body[0]
    if inline:
        expr.value.ctx = ctx
        return expr.value
    else:
        expr.ctx = ctx
        return expr

def new_ast_assign_from_str(s):
    expr = ast.parse(s).body[0]
    return expr

def new_ast_name(name, ctx=None):
    ctx = ast.Load() if ctx == None else ctx
    return ast.Name(id=name, ctx=ctx)

def new_ast_subscript(value, indices, ctx=ast.Load()):
    if len(indices) == 1:
        slice = indices[0]
    else:
        slice = ast.Tuple(elts=indices, ctx=ast.Load())
    return ast.Subscript(value=value, slice=slice, ctx=ctx)

def new_ast_assign(target, value):
    return ast.Assign(targets=[target], value=value, lineno=None)

def new_ast_list(elts):
    return ast.List(elts=elts, ctx=ast.Load())

def new_ast_add(left, right):
    return ast.BinOp(left=left, op=ast.Add(), right=right)

def new_ast_mul(left, right):
    return ast.BinOp(left=left, op=ast.Mult(), right=right)

def new_ast_sub(left, right):
    return ast.BinOp(left=left, op=ast.Sub(), right=right)

def new_ast_div(left, right):
    return ast.BinOp(left=left, op=ast.Div(), right=right)

def new_ast_function_def(name, args, body):
    return ast.FunctionDef(name=name, args=args, body=body, decorator_list=[], lineno=None)

def new_ast_attribute(value, attr, ctx=ast.Load()):
    return ast.Attribute(value=value, attr=attr, ctx=ctx)

def deepcopy_ast_node(node, ctx=None):
    newnode = copy.deepcopy(node)
    newnode.ctx = ctx
    return newnode

def new_ast_return(value):
    return ast.Return(value=value, lineno=None)

def new_ast_arg(name, annotation=None):
    return ast.arg(arg=name, annotation=annotation)

def get_init_value_for_reduction(f):
    if f == 'max':
        return new_ast_const(-float('inf'))
    elif f == 'min':
        return new_ast_const(float('inf'))
    elif f == 'sum' or f == 'matmul':
        return new_ast_const(0)
    else:
        assert False

def load_code(src, keep_file=False):
    from pathlib import Path
    import sys, importlib, hashlib
    module_name = hashlib.sha256(src.encode()).hexdigest()
    file_name = f"/tmp/tmp_{module_name}.py"
    Path(file_name).write_text(src, encoding='utf-8')
    spec = importlib.util.spec_from_file_location(f"module_{module_name}", file_name)
    foo = importlib.util.module_from_spec(spec)
    sys.modules[f"module_{module_name}"] = foo
    spec.loader.exec_module(foo)
    if not keep_file:
        Path(file_name).unlink()
    return foo


import ast
import textwrap
import astpass as at

def test_get_used_names():
    code = """
    for i in range(N):
        c[i] = a[i] + b[i]
    """
    tree = ast.parse(textwrap.dedent(code))
    names = at.get_used_names(tree, no_funcname=False)
    assert names == ['i', 'range', 'N', 'c', 'a', 'b']

    names = at.get_used_names(tree, no_funcname=True)
    assert names == ['i', 'N', 'c', 'a', 'b']



import ast
from astpass.utils import *

class ArrayReferenceCheck(ast.NodeVisitor):
    def __init__(self):
        self.array_indices = {}
        self.array_referenced_times = {}
        self.array_ever_written = {}
        self.always_same_index = {}

    def visit_Subscript(self, node):
        
        if isinstance(node.value, ast.Name):
            dump(node)
            name = node.value.id
            self.array_referenced_times[name] = self.array_referenced_times.get(name, 0) + 1
            if name not in self.array_ever_written:
                self.array_ever_written[name] = False

            # Check if the array is always referenced with the same index
            if name not in self.array_indices:
                self.array_indices[name] = ast.unparse(node.slice)
                self.always_same_index[name] = True
            else:
                if self.array_indices[name] != ast.unparse(node.slice):
                    self.always_same_index[name] = False

            if isinstance(node.ctx, ast.Store):
                self.array_ever_written[name] = True
            
        return node


class IntraloopScalarReplacement(ast.NodeTransformer):
    def visit_For(self, node):
        self.generic_visit(node)

        visitor = ArrayReferenceCheck()
        visitor.visit(node)

        scalar_count = 0
     
        for varname in visitor.always_same_index:
            indices = visitor.array_indices[varname]
            # Only perform the replacement if this condition is met
            if visitor.always_same_index[varname] and visitor.array_referenced_times[varname] > 0:
                scalar_var = f'__scalar_{scalar_count}'
                ReplaceSubscriptsWithName(varname, indices, scalar_var).visit(node)
                scalar_count += 1

                # Insert the stores at the end of the loop if the array is ever written to
                if visitor.array_ever_written[varname]:
                    node.body.append(
                        new_ast_assign_from_str(f'{varname}[{indices}] = {scalar_var}')
                    )

        # # Scan the loop body and replace the subscripts with the generated scalar variables
        # for i, sub in enumerate(candidates):
        #     scalar_var = f'__scalar_{i}'
        #     ReplaceSubscriptsWithName(sub, scalar_var).visit(node)

        # # Insert the stores at the end of the loop
        # for i, sub in enumerate(candidates):
        #     scalar_var = f'__scalar_{i}'
        #     node.body.append(
        #         new_ast_assign(
        #             deepcopy_ast_node(sub, ctx=ast.Store()),
        #             new_ast_name(scalar_var)
        #         )
        #     )

        return node

class ReplaceSubscriptsWithName(ast.NodeTransformer):
    def __init__(self, name, indices, scalar_var):
        self.name = name
        self.indices = indices
        self.scalar_var = scalar_var

    def visit_Subscript(self, node):
        if isinstance(node.value, ast.Name) and node.value.id == self.name and ast.unparse(node.slice) == self.indices:
            if isinstance(node.ctx, ast.Load):
                return new_ast_name(self.scalar_var)
            else:
                return new_ast_name(self.scalar_var, ctx=ast.Store())
        return node

def transform(node):
    return IntraloopScalarReplacement().visit(node)


import ast
from ..utils import *

class BinaryOpToAssign(ast.NodeTransformer):
    var_count = 0

    def __init__(self):
        self.stmts = []

    def get_new_var(self):
        BinaryOpToAssign.var_count += 1
        return '__v%d' % BinaryOpToAssign.var_count

    def visit_Call(self, node):
        self.generic_visit(node)
        node = new_ast_assign(
            target = new_ast_name(self.get_new_var(), ctx = ast.Store()),
            value = node
        )
        self.stmts.append(node)
        return node

    def visit_BinOp(self, node):
        newleft = self.visit(node.left)
        newright = self.visit(node.right)
        # newleft and newright now may be statements
        if isinstance(newleft, ast.Assign):
            node.left = ast.Name(id = newleft.targets[0].id, ctx = ast.Load())
        else:
            node.left = newleft

        if isinstance(newright, ast.Assign):
            node.right = ast.Name(id = newright.targets[0].id, ctx = ast.Load())
        else:
            node.right = newright

        #assign = ast.Assign(targets = [ast.Name(id = self.get_new_var(), ctx = ast.Store())], value = node, lineno = node.lineno, col_offset = node.col_offset)
        assign = new_ast_assign(
            target = new_ast_name(self.get_new_var(), ctx = ast.Store()),
            value = node
        )
        self.stmts.append(assign)
        return assign

    def visit_Compare(self, node):
        assert len(node.comparators) == 1
        newleft = self.visit(node.left)
        newright = self.visit(node.comparators[0])
        # newleft and newright now may be statements
        if isinstance(newleft, ast.Assign):
            node.left = ast.Name(id = newleft.targets[0].id, ctx = ast.Load())
        else:
            node.left = newleft

        if isinstance(newright, ast.Assign):
            node.right = ast.Name(id = newright.targets[0].id, ctx = ast.Load())
        else:
            node.right = newright

        assign = ast.Assign(targets = [ast.Name(id = self.get_new_var(), ctx = ast.Store())], value = node, lineno = node.lineno, col_offset = node.col_offset)
        self.stmts.append(assign)
        return assign


class ToSingleOperatorStmts(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.value, ast.BinOp):
            visitor = BinaryOpToAssign()
            assign = visitor.visit(node.value)
            #node.value = assign.targets[0]
            node.value = ast.Name(id = assign.targets[0].id, ctx = ast.Load())
            return visitor.stmts + [node]
        elif isinstance(node.value, ast.Call):
            visitor = BinaryOpToAssign()
            newargs = [visitor.visit(arg) for arg in node.value.args]
            node.value.args = []
            for newargs in newargs:
                if isinstance(newargs, ast.Assign):
                    node.value.args.append(ast.Name(id = newargs.targets[0].id, ctx = ast.Load()))
                else:
                    node.value.args.append(newargs)
            return visitor.stmts + [node]
        elif isinstance(node.value, ast.Tuple):
            visitor = BinaryOpToAssign()
            newelts = [visitor.visit(arg) for arg in node.value.elts]
            node.value.elts = []
            for newelt in newelts:
                if isinstance(newelt, ast.Assign):
                    node.value.elts.append(ast.Name(id = newelt.targets[0].id, ctx = ast.Load()))
                else:
                    node.value.elts.append(newelt)
            return visitor.stmts + [node]
        else:
            return node

class ReturnExprToStmt(ast.NodeTransformer):
    def visit_Return(self, node):
        if not isinstance(node.value, ast.Name):
            assign = ast.Assign(targets = [ast.Name(id = '__ret', ctx = ast.Store())], value = node.value, lineno = node.lineno, col_offset = node.col_offset)
            node.value = ast.Name(id = '__ret', ctx = ast.Load())
            return [assign] + [node]
        else:
            return node


class RemoveRedundantAssign(ast.NodeTransformer):
    def __init__(self):
        self.prev = None

    def visit_Assign(self, node):
        if isinstance(node.value, ast.Name) and node.value.id.startswith('__v'):
            assert self.prev != None and self.prev.targets[0].id == node.value.id
            self.prev.targets[0] = node.targets[0]
            return
        else:
            self.prev = node
            return node


def transform(tree, **kwargs):
    tree = ReturnExprToStmt().visit(tree)
    tree = ToSingleOperatorStmts().visit(tree)
    tree = RemoveRedundantAssign().visit(tree)
    return tree



import ast

class ReplaceName(ast.NodeTransformer):
    def __init__(self, old_name, new_name):
        self.old_name = old_name
        self.new_name = new_name

    def visit_Name(self, node):
        if node.id == self.old_name:
            node.id = self.new_name
        return self.generic_visit(node)

def transform(node, old_name, new_name):
    return ReplaceName(old_name, new_name).visit(node)


import ast
import textwrap
from astpass.passes import normalize_ranges

def test1():
    code = """
    range(10)
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = normalize_ranges.transform(tree)
    new_code = ast.unparse(new_tree)

    expected = """
    range(0, 10, 1)
    """
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test2():
    code = """
    range(1, 10)
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = normalize_ranges.transform(tree)
    new_code = ast.unparse(new_tree)

    expected = """
    range(1, 10, 1)
    """
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test3():
    code = """
    range(1, 10, 2)
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = normalize_ranges.transform(tree)
    new_code = ast.unparse(new_tree)

    expected = """
    range(1, 10, 2)
    """
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))


import ast

class AddFuncDecorator(ast.NodeTransformer):
    def __init__(self, decorator: str):
        if decorator.startswith('@'):
            decorator = decorator[1:]
        self.decorator = decorator

    def visit_FunctionDef(self, node):
        dec_node = ast.parse(self.decorator).body[0]
        assert isinstance(dec_node, ast.Expr)
        node.decorator_list.append(dec_node.value)
        return node

def transform(node, decorator):
    return AddFuncDecorator(decorator).visit(node)



import ast

def is_call(node, names=None):
    '''
    Check if the node is a call node. If `names` is not None, check if the function name is in `names`.
    '''
    if isinstance(node, ast.Call):
        if names:
            names = names if isinstance(names, (tuple, list)) else (names,)
            return ast.unparse(node.func) in names
        else:
            return True
    else:
        return False
    
def str_to_ast_expr(expr_str):
    return ast.parse(expr_str).body[0].value



import ast
import textwrap
from astpass.passes import hoist_shape_access

def test1():
    code = """
    for i in range(a.shape[0]):
        x = a.shape[0] + b.shape[1]
        y = a[i] + b[i]
    """
    expected = """
    a_shape_0 = a.shape[0]
    b_shape_1 = b.shape[1]
    for i in range(a_shape_0):        
        x = a_shape_0 + b_shape_1
        y = a[i] + b[i]
    """
    tree = ast.parse(textwrap.dedent(code))
    tree = hoist_shape_access.transform(tree)
    assert ast.dump(tree) == ast.dump(ast.parse(textwrap.dedent(expected)))

def test2():
    code = """
    for i in range(a.shape[0]):
        for j in range(a.shape[1]):
            b[i,j] = a[i,j] + 1
    """
    expected = """
    a_shape_0 = a.shape[0]
    a_shape_1 = a.shape[1]
    for i in range(a_shape_0):
        for j in range(a_shape_1):
            b[i,j] = a[i,j] + 1
    """
    tree = ast.parse(textwrap.dedent(code))
    tree = hoist_shape_access.transform(tree)
    print(ast.unparse(tree))
    assert ast.dump(tree) == ast.dump(ast.parse(textwrap.dedent(expected)))




import ast
import textwrap
import numpy as np

from astpass.passes import vector_op_to_loop

def test_add1():
    code = """
    c = a + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': 1.0,
        'c': np.empty(10)
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    for __i0 in range(0, 10):
        c[__i0] = a[__i0] + b
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_add2():
    code = """
    c = a[:] + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': 1.0,
        'c': np.empty(10)
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    for __i0 in range(0, 10):
        c[__i0] = a[__i0] + b
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_add3():
    code = """
    c[:] = a[:] + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': 1.0,
        'c': np.empty(10)
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    for __i0 in range(0, 10):
        c[__i0] = a[__i0] + b
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_add4():
    code = """
    c[:] = a[:] * 2 + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': 1.0,
        'c': np.empty(10)
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    for __i0 in range(0, 10):
        c[__i0] = a[__i0] * 2 + b
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_np_sum1():
    code = """
    c = np.sum(a)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'c': 0.0,
        'np': np
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    c = 0
    for __i0 in range(0, 10):
        c = c + a[__i0]
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_np_sum2():
    code = """
    c = np.sum(a * b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': np.random.randn(10),
        'c': 0.0,
        'np': np
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    c = 0
    for __i0 in range(0, 10):
        c = c + a[__i0] * b[__i0]
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_np_max1():
    code = """
    c = np.max(a)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'c': 0.0,
        'np': np
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    c = float('-inf')
    for __i0 in range(0, 10):
        c = max(c, a[__i0])
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test_np_min1():
    code = """
    c = np.min(a)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'c': 0.0,
        'np': np
    }
    tree = vector_op_to_loop.transform(tree, rt_vals)

    expected = """
    c = float('inf')
    for __i0 in range(0, 10):
        c = min(c, a[__i0])
    """
    new_code = ast.unparse(tree)
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))


import ast

class RemoveFuncDecorator(ast.NodeTransformer):
    def visit_FunctionDef(self, node):
        node.decorator_list = []
        return node

def transform(node):
    return RemoveFuncDecorator().visit(node)


import pkgutil
from pathlib import Path

body = ""
for modinfo in pkgutil.iter_modules([f'{Path(__file__).parent}/passes']):
    name = modinfo.name
    body += f"def {name}(tree, *args):\n"
    body += f"    '''Apply the `{name}` AST transform.'''\n"
    body += f"    from .passes import {name} as m\n"
    body += f"    return m.transform(tree, *args)\n"
    body += "\n"

Path(f'{Path(__file__).parent}/__init__.py').write_text(body)


import ast

class WhereToTernary(ast.NodeTransformer):
    def visit_Call(self, node):
        func = node.func
        if (
            (isinstance(func, ast.Name) and func.id == 'where')
            or 
            (isinstance(func, ast.Attribute)
             and isinstance(func.value, ast.Name)
             and func.attr == 'where'
             and func.value.id in ['np', 'numpy', 'torch', 'cupy'])
        ):
            assert len(node.args) == 3, f"where should have 3 arguments, but got {len(node.args)}"
            return ast.IfExp(
                test=node.args[0],
                body=node.args[1],
                orelse=node.args[2],
                lineno=node.lineno
            )
        return node

def transform(node):
    return WhereToTernary().visit(node)


import ast

class RemoveFuncArgAnnotation(ast.NodeTransformer):
    def visit_arguments(self, node):
        for arg in node.args:
            arg.annotation = None
        return node

def transform(node):
    return RemoveFuncArgAnnotation().visit(node)


import ast
import textwrap
from astpass.passes import add_func_decorator

def test1():
    code = """
    def foo():
        return 42
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = add_func_decorator.transform(tree, "jit")
    new_code = ast.unparse(new_tree)

    expected = """
    @jit
    def foo():
        return 42
    """
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))

def test2():
    code = """
    def foo():
        return 42
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = add_func_decorator.transform(tree, "numba.njit(parallel=True)")
    new_code = ast.unparse(new_tree)

    expected = """
    @numba.njit(parallel=True)
    def foo():
        return 42
    """
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))



import ast

class GetUsedNames(ast.NodeVisitor):
    def __init__(self, no_funcname):
        self.used = []
        self.no_funcname = no_funcname

    def visit_Name(self, node):
        if node.id not in self.used:
            self.used.append(node.id)

    def visit_Call(self, node: ast.Call):
        if self.no_funcname:
            for arg in node.args:
                self.visit(arg)
        else:
            self.generic_visit(node)

def analyze(tree, no_funcname):
    visitor = GetUsedNames(no_funcname)
    visitor.visit(tree)
    return visitor.used


def uop_generic(a):
    return a

def binop_generic(left, right):
    if left == right:
        return left
    elif left == ():
        return right
    elif right == ():
        return left
    else:
        raise NotImplementedError("Shape broadcasting is not implemented")
    
def compare_generic(left, right):
    return binop_generic(left, right)

def ifexp_generic(test, body, orelse):
    assert test == () and body == orelse
    return body
    
def matmul_generic(left, right):
    if not (len(left) > 0 and len(right) > 0):
        raise RuntimeError("Matmul cannot happen on scalar operands")

    if left[-1] != right[0]:
        raise RuntimeError(f"Mismatched contracting dimension found for matmul: {left[-1]} and {right[0]}")
    return left[:-1] + right[1:]

def range(*args):
    '''
    The shape of range cannot be determined by the shape of its arguments.
    So simply return a None here, need another pass to pass the values of 
    the arguments.
    '''
    return None 

def slice(low, up, step):
    # ---- Validate inputs ----
    for name, arg in [("low", low), ("up", up), ("step", step)]:
        if not isinstance(arg, (int, type(None), str)):
            raise TypeError(f"{name} must be int, str, or None")

    step = 1 if step is None else step
    if step != 1:
        raise RuntimeError("Non-1 step is not supported in slice")

    low = 0 if low is None else low

    if isinstance(low, int) and low < 0:
        raise ValueError("Slice lower bound must be non-negative")

    # ---- Dispatch rules ----
    table = {
        (int, int): lambda low, up: (up - low,),
        (int, type(None)): lambda low, up: (None,) if low == 0 else (-low,),
        (int, str): lambda low, up: (f":{up}",) if low == 0 else (f"{low}:{up}",),
        (str, int): lambda low, up: (f"{low}:{up}",),
        (str, type(None)): lambda low, up: (f"{low}:",),
        (str, str): lambda low, up: (f"{low}:{up}",),
    }

    key = (type(low), type(up))
    if key not in table:
        raise NotImplementedError(f"Unsupported slice shape: low={low}, up={up}")

    return table[key](low, up)

def subscript(base, indices):
    shape = []
    for i, idx in enumerate(indices):
        # Case 1: scalar index
        if idx == ():
            pass
        elif len(idx) == 1:
            size = idx[0]
            if not isinstance(size, (int, type(None), str)):
                raise TypeError("A shape dimension must be int, str, or None")
            
            table = {
                int: lambda size: size if size >= 0 else base[i] + size,
                type(None): lambda size: base[i],
                str: lambda size: size
            }

            key = type(size)
            shape.append(table[key](size))
        else:
            assert False, "Should not reach here"
    shape += base[len(indices):]
    return tuple(shape)

def numpy_sin(a):
    return uop_generic(a)

def numpy_cos(a):
    return uop_generic(a)

def numpy_tan(a):
    return uop_generic(a)

def numpy_sinh(a):
    return uop_generic(a)

def numpy_cosh(a):
    return uop_generic(a)

def numpy_tanh(a):
    return uop_generic(a)

def numpy_round(a, decimals=()):
    return uop_generic(a)

def numpy_rint(a):
    return uop_generic(a)

def numpy_log(a):
    return uop_generic(a)

def numpy_exp(a):
    return uop_generic(a)

def numpy_sqrt(a):
    return uop_generic(a)

def numpy_pow(a, b):
    assert len(b) == 0 or len(b) == 1 or len(b) == len(a)
    return a

def numpy_power(a, b):
    return numpy_pow(a, b)

def numpy_add(a, b):
    return binop_generic(a, b)

def numpy_subtract(a, b):
    return binop_generic(a, b)

def numpy_multiply(a, b):
    return binop_generic(a, b)

def numpy_divide(a, b):
    return binop_generic(a, b)

def numpy_minimum(a, b):
    return binop_generic(a, b)

def numpy_maximum(a, b):
    return binop_generic(a, b)

def numpy_reduce_generic(a, axis=None):
    if axis != None:
        assert isinstance(axis, int)
        if not (axis >= 0 and axis < len(a)):
            raise ValueError(f"axis={axis} must be in range [0, {len(a)})")
        
        a = list(a)
        del a[axis]
        return tuple(a)
    else:
        return ()

## Built-in functions
def pow(a, b):
    return numpy_pow(a, b)

def min(a, b):
    assert a == () and b == ()
    return ()

def max(a, b):
    assert a == () and b == ()
    return ()

def erf(a):
    return uop_generic(a)


import ast
import textwrap
from astpass.passes import shape_analysis
import numpy as np

def test_unary1():
    code = """
    -1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('1', ()), ('-1', ())]

def test_unary2():
    code = """
    -1.0
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('1.0', ()), ('-1.0', ())]

def test_unary3():
    code = """
    -a
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": 1}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('-a', ())]

def test_unary4():
    code = """
    -a
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(10)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (10,)), ('-a', (10,))]

def test_binary1():
    code = """
    1 + 1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('1', ()), ('1', ()), ('1 + 1', ())]

def test_binary2():
    code = """
    1.0 + 1.0
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('1.0', ()), ('1.0', ()), ('1.0 + 1.0', ())]

def test_binary3():
    code = """
    a + 1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": 1}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('1', ()), ('a + 1', ())]

def test_binary4():
    code = """
    a + 1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.int32(1)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('1', ()), ('a + 1', ())]

def test_binary5():
    code = """
    a + 1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('1', ()), ('a + 1', (100,))]

def test_binary6():
    code = """
    a + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('a + b', (100,))]

def test_binary7():
    code = """
    a + b * 2 - 1
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('2', ()), ('b * 2', (100,)), \
        ('a + b * 2', (100,)), ('1', ()), ('a + b * 2 - 1', (100,))]
    
def test_binary8():
    code = """
    a / 2 - b // 2
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('2', ()), ('a / 2', (100,)), \
        ('b', (100,)), ('2', ()), ('b // 2', (100,)), ('a / 2 - b // 2', (100,))]

def test_binary9():
    code = """
    a @ b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('a @ b', ())]

def test_binary10():
    code = """
    a @ b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(10, 100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (10, 100,)), ('b', (100,)), ('a @ b', (10,))]

def test_compare1():
    code = """
    a < b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": np.random.randn(100)}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('a < b', (100,))]

def test_compare2():
    code = """
    a > b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {"a": np.random.randn(100), "b": 0}
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', ()), ('a > b', (100,))]
    
def test_call_np_add1():
    code = """
    np.add(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100), 
        "b": np.random.randn(100),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('np.add(a, b)', (100,))]

def test_call_pow1():
    code = """
    pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": 2, 
        "b": 3,
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('b', ()), ('pow(a, b)', ())]

def test_call_pow2():
    code = """
    pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": 3,
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', ()), ('pow(a, b)', (100,))]

def test_call_pow3():
    code = """
    pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": np.ones(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('pow(a, b)', (100,))]

def test_call_np_pow1():
    code = """
    np.pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": 2, 
        "b": 3,
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('b', ()), ('np.pow(a, b)', ())]

def test_call_np_pow2():
    code = """
    np.pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": 3,
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', ()), ('np.pow(a, b)', (100,))]

def test_call_np_pow3():
    code = """
    np.pow(a, b)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": np.ones(100),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('b', (100,)), ('np.pow(a, b)', (100,))]

def test_call_np_sum1():
    code = """
    np.sum(a)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (100,)), ('np.sum(a)', ())]

def test_call_np_sum2():
    code = """
    np.sum(a)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(3, 4),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (3, 4)), ('np.sum(a)', ())]

def test_call_np_sum3():
    code = """
    np.sum(a, 0)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(3, 4),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if shape != ()]
    assert results == [('a', (3, 4)), ('np.sum(a, 0)', (4,))]

def test_call_np_sum4():
    code = """
    np.sum(a, 1)
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(3, 4),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if shape != ()]
    assert results == [('a', (3, 4)), ('np.sum(a, 1)', (3,))]

def test_call_np_sum5():
    code = """
    np.sum(a[:, 0])
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(3, 4),
        "np": np
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, (ast.Subscript, ast.Call))]
    assert results == [('a[:, 0]', (3,)), ('np.sum(a[:, 0])', ())]

def test_slice1():
    code = """
    a[0:2]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [('0:2', (2,))]

def test_slice2():
    code = """
    a[:2]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [(':2', (2,))]

def test_slice3():
    code = """
    a[:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [(':', (None,))]

def test_slice4():
    code = """
    a[0:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [('0:', (None,))]

def test_slice5():
    code = """
    a[1:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [('1:', (-1,))]

def test_slice6():
    code = """
    a[1:-1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [('1:-1', (-2,))]

def test_slice7():
    code = """
    a[:-1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [(':-1', (-1,))]

def test_slice8():
    code = """
    a[:i]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Slice)]
    assert results == [(':i', (':i',))]

def test_subscript1():
    code = """
    a[0:2]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[0:2]', (2,))]

def test_subscript2():
    code = """
    a[:2]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[:2]', (2,))]

def test_subscript3():
    code = """
    a[:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[:]', (100,))]

def test_subscript4():
    code = """
    a[0:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[0:]', (100,))]

def test_subscript5():
    code = """
    a[1:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[1:]', (99,))]

def test_subscript6():
    code = """
    a[1:-1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[1:-1]', (98,))]

def test_subscript7():
    code = """
    a[:-1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[:-1]', (99,))]

def test_subscript8():
    code = """
    a[:i+1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[:i + 1]', (':i + 1',))]

def test_subscript9():
    code = """
    a[0:3, :, :i+1]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(10, 10, 10),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, ast.Subscript)]
    assert results == [('a[0:3, :, :i + 1]', (3, 10, ':i + 1',))]

def test_subscripted_matmul1():
    code = """
    a[:] @ b[:]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, (ast.Subscript, ast.BinOp))]
    assert results == [('a[:]', (100,)), ('b[:]', (100,)), ('a[:] @ b[:]', ())]

def test_subscripted_matmul2():
    code = """
    a[:i] @ b[:i]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(100),
        "b": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, (ast.Subscript, ast.BinOp))]
    assert results == [('a[:i]', (':i',)), ('b[:i]', (':i',)), ('a[:i] @ b[:i]', ())]

def test_subscripted_matmul3():
    code = """
    a[:i, :j] @ b[:j]
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        "a": np.random.randn(10, 100),
        "b": np.random.randn(100),
    }
    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items() if isinstance(node, (ast.Subscript, ast.BinOp))]
    assert results == [('a[:i, :j]', (':i', ':j')), ('b[:j]', (':j',)), ('a[:i, :j] @ b[:j]', (':i',))]

def test_assign1():
    code = """
    c = a + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': 1,
        'b': 2
    }

    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', ()), ('b', ()), ('a + b', ()), ('c', ())]

def test_assign2():
    code = """
    c = a + b
    """
    tree = ast.parse(textwrap.dedent(code))
    rt_vals = {
        'a': np.random.randn(10),
        'b': 2
    }

    shape_info = shape_analysis.analyze(tree, rt_vals)
    results = [(ast.unparse(node), shape) for node, shape in shape_info.items()]
    assert results == [('a', (10,)), ('b', ()), ('a + b', (10,)), ('c', (10,))]


# def test12():
#     code = """
#     c = a + b
#     """
#     tree = ast.parse(textwrap.dedent(code))
#     rt_vals = {
#         'a': np.random.randn(100),
#         'b': np.random.randn(100)
#     }

#     shape_info = shape_analysis.analyze(tree)
#     for node, shape in shape_info.items():
#         assert shape == (100,)


# def test13():
#     code = """
#     c = a[i] + b
#     """
#     tree = ast.parse(textwrap.dedent(code))
#     rt_vals = {
#         'a': np.random.randn(100),
#         'b': 1.0,
#         'i': 0
#     }

#     shape_info = shape_analysis.analyze(tree)
#     for node, shape in shape_info.items():
#         if isinstance(node, ast.Name) and node.id == 'a':
#             assert shape == (100,)
#         else:
#             assert shape == ()


import ast

class NormalizeRange(ast.NodeTransformer):
    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id == 'range':
            if len(node.args) == 1:
                up = node.args[0]
                node.args = [
                    ast.Constant(0),
                    up,
                    ast.Constant(1)
                ]
            elif len(node.args) == 2:
                low, up = node.args
                node.args = [
                    low,
                    up,
                    ast.Constant(1)
                ]

            assert len(node.args) == 3, ast.dump(node)
        return node
    
def transform(node):
    return NormalizeRange().visit(node)


from setuptools import setup, find_packages

setup(
    name="astpass",                     # Name of the package
    version="0.1.1",                    # Version
    author="Tong Zhou",                 # Your name
    author_email="zt9465@gmail.com", # Your email
    description="A collection of AST-based analysis and transformation passes", # Short description
    long_description=open("README.md").read(), # Long description from README file
    long_description_content_type="text/markdown", # Type of the long description
    url="https://github.com/habanero-lab/astpass", # URL of your project repository
    packages=find_packages(),           # Automatically find the packages in your project
    classifiers=[
        "Programming Language :: Python :: 3",  # Classifiers help users find your project
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',            # Minimum Python version requirement
    install_requires=['ast_comments'],
    extras_require={
        "test": ["numpy", "pytest"],
    },
)



import ast
from ..utils import *

class NameVistor(ast.NodeVisitor):
    def __init__(self):
        self.vars = []

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            self.vars.append(node.id)

    def visit_Call(self, node):
        # Skip visiting the function name and only visit the argument names        
        for arg in node.args:
            self.visit(arg)  # Visit positional arguments
        for kw in node.keywords:
            if kw.value is not None:
                self.visit(kw.value)  # Visit keyword argument values


class AttachDefUseVars(ast.NodeTransformer):
    def visit_Assign(self, node):
        assert isinstance(node.targets[0], ast.Name)
        node.def_vars = [node.targets[0].id]
        visitor = NameVistor()
        visitor.visit(node.value)
        node.use_vars = visitor.vars
        return node

    def visit_Return(self, node):
        visitor = NameVistor()
        visitor.visit(node.value)
        node.use_vars = visitor.vars
        self.generic_visit(node)
        return node

    def visit_While(self, node):
        '''
        In the CFG, the while node only reprents its conditional part
        '''
        visitor = NameVistor()
        visitor.visit(node.test)
        node.use_vars = visitor.vars
        self.generic_visit(node)
        return node

    def visit_If(self, node):
        visitor = NameVistor()
        visitor.visit(node.test)
        node.use_vars = visitor.vars
        self.generic_visit(node)
        return node

    def visit_Expr(self, node):
        visitor = NameVistor()
        visitor.visit(node.value)
        node.use_vars = visitor.vars
        self.generic_visit(node)
        return node

def transform(tree):
    visitor = AttachDefUseVars()
    tree = visitor.visit(tree)
    return tree



import ast
from ...passes.ast_utils import str_to_ast_expr
from ...passes import shape_analysis

class CollectNonzeroShapes(ast.NodeVisitor):
    def __init__(self, shape_info):
        self.shape_info = shape_info
        self.nonzero_shapes = []

    def get_node_shape(self, node):
        if node not in self.shape_info:
            raise KeyError(f"Shape info not found for node {type(node)}: {ast.unparse(node)}")
        return self.shape_info[node]

    def visit_Subscript(self, node):
        shape = self.get_node_shape(node)
        assert shape is not None
        if len(shape) > 0:
            self.nonzero_shapes.append(shape)

    def visit_Call(self, node):
        for arg in node.args:
            self.visit(arg)

    def visit_Name(self, node):
        shape = self.get_node_shape(node)
        assert shape is not None
        if len(shape) > 0:
            self.nonzero_shapes.append(shape)

class Scalarize(ast.NodeTransformer):
    def __init__(self, shape_info, idx):
        self.shape_info = shape_info
        self.idx = idx

    def get_node_shape(self, node):
        if node not in self.shape_info:
            raise KeyError(f"Shape info not found for node {type(node)}: {ast.unparse(node)}")
        return self.shape_info[node]
    
    def visit_Call(self, node):
        node.args = [self.visit(arg) for arg in node.args]
        return node

    def visit_Name(self, node):
        shape = self.get_node_shape(node)
        if len(shape) > 0:
            assert len(shape) == 1
            return ast.Subscript(
                value=ast.Name(id=node.id, ctx=ast.Load()),
                slice=ast.Name(id=self.idx, ctx=ast.Load()),
                ctx=ast.Load()
            )
        else:
            return node
        
    def visit_Subscript(self, node):
        shape = self.get_node_shape(node)
        if len(shape) == 0:
            return node
        
        indices = node.slice.elts if isinstance(node.slice, ast.Tuple) else (node.slice,)
        num_slices = sum([isinstance(idx, ast.Slice) for idx in indices])
        assert num_slices in [0, 1], f"A subscript should have 0 or 1 sliced indices, but got {num_slices}"

        new_scalar_index = ast.Name(id=self.idx, ctx=ast.Load())
        if num_slices == 0:
            new_indices = indices + (new_scalar_index,)
        elif num_slices == 1:
            new_indices = [new_scalar_index if isinstance(idx, ast.Slice) else idx for idx in indices]

        return ast.Subscript(
                value=node.value,
                slice=ast.Tuple(elts=new_indices, ctx=ast.Load()) if len(new_indices) > 1 else new_indices[0],
                ctx=node.ctx
            )


class PointwiseExprToLoop(ast.NodeTransformer):
    def __init__(self, shape_info, loop_index_prefix=None):
        self.shape_info = shape_info
        self.loop_index_prefix = loop_index_prefix if loop_index_prefix is not None else "__i"
        self.loop_index_count = 0

    def get_node_shape(self, node):
        if node not in self.shape_info:
            raise KeyError(f"Shape info not found for node {type(node)}: {ast.unparse(node)}")
        return self.shape_info[node]

    def get_new_loop_index(self):
        name = f"{self.loop_index_prefix}{self.loop_index_count}"
        self.loop_index_count += 1
        return name

    def get_loop_bounds(self, shapes):
        if not all([s == shapes[0] for s in shapes]):
            raise RuntimeError(f"Shapes are not the same: {shapes}")
        
        if not all([len(s) == 1 for s in shapes]):
            raise RuntimeError(f"Only 1D array expansion is supported, but got shapes: {shapes}")
        
        bound = shapes[0][0]
        assert isinstance(bound, (int, str))
        if isinstance(bound, int):
            low, up = 0, bound
        elif isinstance(bound, str):
            low, up = bound.split(":")
            assert up != ''
            low = 0 if low == '' else low
        return low, up

    def visit_Assign(self, node):        
        shape_visitor = CollectNonzeroShapes(self.shape_info)
        shape_visitor.visit(node)
        nonzero_shapes = shape_visitor.nonzero_shapes
        if nonzero_shapes:
            low, up = self.get_loop_bounds(nonzero_shapes)
            return self.gen_loop(node, low, up)
        else:
            return node

    def gen_loop(self, node, low, up):
        index = self.get_new_loop_index()
        loop = ast.For(
            target=ast.Name(id=index, ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(id='range', ctx=ast.Load()),
                args=[
                    str_to_ast_expr(low) if isinstance(low, str) else ast.Constant(low),
                    str_to_ast_expr(up) if isinstance(up, str) else ast.Constant(up)
                ],
                keywords=[]
            ),
            body=[Scalarize(self.shape_info, index).visit(node)],
            orelse=[],
            lineno=node.lineno
        )
        return loop

def transform(tree, runtime_vals, loop_index_prefix=None):
    '''
    This pass detects and rewrites tensor expressions to explicit loops.

    Note 
    ----
    This pass assumes that all arrays have been allocated, and it only generates
    the loops, no memory allocations will be performed. In other words, all variables
    appeared in the input code should already be defined.
    '''
    shape_info = shape_analysis.analyze(tree, runtime_vals)
    return PointwiseExprToLoop(shape_info, loop_index_prefix).visit(tree)


def add_func_decorator(tree, decorator):
    """
    Adds a decorator to all functions in the AST.

    Parameters
    ----------
    tree : ast.AST
        The AST of the Python code to transform.
    decorator : str
        The decorator to add. Can be a fully-qualified name
        (e.g., "numba.jit") or a simple name (e.g., "jit").

    Returns
    -------
    ast.AST
        The transformed AST.
    """
    from .passes import add_func_decorator as m
    return m.transform(tree, decorator)

def remove_func_decorator(tree):
    """
    Remove all decorators from functions in the AST.

    Parameters
    ----------
    tree : ast.AST
        The AST of the Python code to transform.

    Returns
    -------
    ast.AST
        The transformed AST.
    """
    from .passes import remove_func_decorator as m
    return m.transform(tree)

def get_used_names(tree, no_funcname=True):
    '''
    Return a sorted list of all names used in the AST.

    Parameters
    ----------
    tree : ast.AST
        The AST of the Python code to analyze.
    no_funcname : bool, optional
        If True, function names are excluded from the result. Default is True.

    Returns
    -------
    list of str
        The sorted list of used names.
    '''
    from .passes import get_used_names as m
    return m.analyze(tree, no_funcname)

def hoist_shape_attr(tree):
    '''
    Hoists shape attribute accesses in the AST. For example, 
    `a.shape[0]` is replaced with `a_shape_0`, and an assignment
    `a_shape_0 = a.shape[0]` is inserted before the loop.

    Parameters
    ----------
    tree : ast.AST
        The AST of the Python code to transform.

    Returns
    -------
    ast.AST
        The transformed AST.
    '''
    from .passes import hoist_shape_attr as m
    return m.transform(tree)


import ast
import inspect
from . import func_table
from ..ast_utils import is_call

class AnalyzeExprShapes(ast.NodeVisitor):
    def __init__(self, rt_vals):
        self.node_shapes = {}
        self.var_shapes = {}
        self.modules = {}
        self.init_rt_var_shapes(rt_vals)
        self.init_module_names(rt_vals)

    def init_rt_var_shapes(self, rt_vals):
        for var, val in rt_vals.items():
            if isinstance(val, (int, float, bool)):
                self.var_shapes[var] = ()
            elif hasattr(val, 'shape'):
                self.var_shapes[var] = val.shape
            else:
                raise RuntimeError(f"Unsupported type: {type(val)}")

    def init_module_names(self, rt_vals):
        for var, val in rt_vals.items():
            if inspect.ismodule(val):
                self.modules[var] = val

    # Two types of leaf nodes
    def visit_Constant(self, node):
        if isinstance(node.value, (int, float, bool)):
            self.node_shapes[node] = ()
        else:
            raise RuntimeError(f"Unsupported constant type: {type(node.value)}")

    def visit_Name(self, node):
        if node.id in self.var_shapes:
            self.node_shapes[node] = self.var_shapes[node.id]
        else:
            raise RuntimeError(f"Name {node.id} not found in runtime values")
    
    # Ways to combine nodes
    def visit_UnaryOp(self, node):
        self.generic_visit(node)
        f = getattr(func_table, 'uop_generic')
        self.node_shapes[node] = f(self.node_shapes[node.operand])
    
    def visit_BinOp(self, node):
        self.generic_visit(node)
        if isinstance(node.op, (ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.BitAnd, ast.BitOr, ast.BitXor)):
            f = getattr(func_table, 'binop_generic')
            self.node_shapes[node] = f(self.node_shapes[node.left], self.node_shapes[node.right])
        elif isinstance(node.op, ast.MatMult):
            f = getattr(func_table, 'matmul_generic')
            self.node_shapes[node] = f(self.node_shapes[node.left], self.node_shapes[node.right])
        else:
            raise NotImplementedError(f"Binary operator {node.op} not implemented")
        
    def visit_Compare(self, node):
        self.generic_visit(node)
        assert len(node.comparators) == 1
        f = getattr(func_table, 'compare_generic')        
        self.node_shapes[node] = f(self.node_shapes[node.left], self.node_shapes[node.comparators[0]])

    def visit_IfExp(self, node: ast.IfExp):
        self.generic_visit(node)
        f = getattr(func_table, 'ifexp_generic')
        self.node_shapes[node] = f(self.node_shapes[node.test], self.node_shapes[node.body], self.node_shapes[node.orelse])

    def dispatch_call(self, f_name, args):
        if f_name in ['numpy_sum', 'numpy_min', 'numpy_max', 'numpy_argmin', 'numpy_argmax']:
            assert len(args) in [1, 2], f"numpy_<reduce> should either one or two arguments, but got {len(args)}"
            func_args = [self.node_shapes[args[0]]]
            if len(args) == 2:
                if not (isinstance(args[1], ast.Constant) and isinstance(args[1].value, int)):
                    raise RuntimeError("Second argument for numpy_<reduce> should be an int constant for shape analysis")
                
                func_args.append(args[1].value)
            return func_table.numpy_reduce_generic(*func_args)
        else:
            f = getattr(func_table, f_name)
            return f(*[self.node_shapes[arg] for arg in args])
        
    def visit_Call(self, node: ast.Call):
        for arg in node.args:
            self.visit(arg)

        if not (
            (isinstance(node.func, ast.Name) or 
             isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name))
        ):
            raise RuntimeError("Function calls only suport named calls or named module calls")

        if isinstance(node.func, ast.Name):
            f_name = node.func.id
        elif isinstance(node.func, ast.Attribute):
            if not node.func.value.id in self.modules:
                raise KeyError(f"Module {node.func.value.id} not found in runtime vals")
            
            module_name = self.modules[node.func.value.id].__name__
            f_name = f"{module_name}_{node.func.attr}"
        else:
            assert False, "Impossible path"

        self.node_shapes[node] = self.dispatch_call(f_name, node.args)

    def visit_Subscript(self, node):
        self.generic_visit(node)
        indices = []
        if isinstance(node.slice, ast.Tuple):
            for index in node.slice.elts:
                indices.append(self.node_shapes[index])
        else:
            indices.append(self.node_shapes[node.slice])
        f = getattr(func_table, 'subscript')
        self.node_shapes[node] = f(self.node_shapes[node.value], indices)

    def visit_Slice(self, node: ast.Slice):  
        if (
            isinstance(node.upper, ast.UnaryOp)
            and isinstance(node.upper.op, ast.USub) 
            and isinstance(node.upper.operand, ast.Constant)
        ):
            node.upper = ast.Constant(value=-node.upper.operand.value)

        args = []
        for arg in [node.lower, node.upper, node.step]:
            if arg is None:
                args.append(None)
            elif isinstance(arg, ast.Constant):
                args.append(arg.value)
            elif isinstance(arg, ast.expr):
                args.append(ast.unparse(arg))
            else:
                raise RuntimeError("Should not reach here")
        
        f = getattr(func_table, 'slice')
        self.node_shapes[node] = f(*args)

class AnalyzeAssignShapes(AnalyzeExprShapes):
    def __init__(self, rt_vals):
        super().__init__(rt_vals)

    def visit_Assign(self, node):
        self.visit(node.value)
        target = node.targets[0]
        if isinstance(target, ast.Name) and target.id not in self.var_shapes:
            self.var_shapes[target.id] = self.node_shapes[node.value]

        self.visit(target)
        # Check if the shape of the target and the value are the same
        if self.node_shapes[target] != self.node_shapes[node.value]:
            raise RuntimeError(f"Shapes mismatch for assignment: {ast.unparse(node)}")
        
    def visit_For(self, node):
        # Only for-range calls are supported
        assert is_call(node.iter, ["range"])
        self.var_shapes[node.target.id] = ()
        self.generic_visit(node)


def analyze(tree, rt_vals):
    visitor = AnalyzeAssignShapes(rt_vals)
    visitor.visit(tree)
    return visitor.node_shapes


import ast
import textwrap
import astpass as at

def test_remove_func_decorator_simple():
    code = """
    @decorator
    def foo():
        return 42
    """
    tree = ast.parse(textwrap.dedent(code))
    new_tree = at.remove_func_decorator(tree)
    new_code = ast.unparse(new_tree)

    expected = """
    def foo():
        return 42
    """
    # Normalize whitespace for comparison
    assert new_code == ast.unparse(ast.parse(textwrap.dedent(expected)))



import ast
from .. import shape_analysis
from ...passes.ast_utils import is_call, str_to_ast_expr
from .convert_point_wise import PointwiseExprToLoop, Scalarize

class ReductionAndPWExprToLoop(PointwiseExprToLoop):
    def get_reduce_op(self, call_node: ast.Call):
        func = ast.unparse(call_node.func)
        table = {
            'sum': 'sum',
            'min': 'min',
            'max': 'max',
            'np.sum': 'sum',
            'np.min': 'min',
            'np.max': 'max',
            'torch.sum': 'sum',
            'torch.min': 'min',
            'torch.max': 'max',
        }
        return table[func]
    
    def gen_initialization(self, reduce_op, var):
        value = None
        if reduce_op == 'sum':
            value = ast.Constant(0.0)
        elif reduce_op == 'max':
            value = str_to_ast_expr("float('-inf')")
        elif reduce_op == 'min':
            value = str_to_ast_expr("float('inf')")
        else:
            raise NotImplementedError
        
        return ast.Assign(
            targets=[ast.Name(id=var, ctx=ast.Store())],
            value=value,
            lineno=None
        )
    
    def rewrite_reduction_assign(self, reduce_op, var, orig_value):
        value = None
        if reduce_op == 'sum':
            value=ast.BinOp(
                op=ast.Add(),
                left=ast.Name(id=var, ctx=ast.Load()),
                right=orig_value.args[0]
            )
        elif reduce_op == 'max' or reduce_op == 'min':
            value=ast.Call(
                func=ast.Name(id=reduce_op, ctx=ast.Load()),
                args=[
                    ast.Name(id=var, ctx=ast.Load()),
                    orig_value.args[0]
                ],
                keywords=[]
            )
        else:
            raise NotImplementedError
        
        return ast.Assign(
            targets=[ast.Name(id=var, ctx=ast.Store())],
            value=value,
            lineno=None
        )
    
    def is_reduction_call(self, node):
        return is_call(node, [
            "np.sum", "np.min", "np.max",
            "torch.sum", "torch.min", "torch.max",
            "sum", "min", "max"
            ]
        )

    def get_temp_reduction_var(self, reduce_op):
        return f"__reduce_{reduce_op}_var"

    def gen_loop(self, node: ast.Assign, low: int|str, up: int|str):
        loop = super().gen_loop(node, low, up)
        # A convenient attribute for APPy
        loop._simd_okay = True
        if self.is_reduction_call(node.value):
            reduce_op = self.get_reduce_op(node.value)
            if len(self.get_node_shape(node.targets[0])) > 0:
                raise RuntimeError(f"Only 1D array reduction is supported, but got target: {ast.dump(node.targets[0])}")
            var = self.get_temp_reduction_var(reduce_op)
            init_stmt = self.gen_initialization(reduce_op, var)
            loop.body = [self.rewrite_reduction_assign(reduce_op, var, node.value)]
            reassign_stmt = ast.Assign(
                targets=[node.targets[0]],
                value=ast.Name(id=var, ctx=ast.Load()),
                lineno=node.lineno
            )
            # A convenient attribute for APPy
            loop._reduction = (reduce_op, var)
            return init_stmt, loop, reassign_stmt
        else:
            return loop
    
def transform(tree, runtime_vals, loop_index_prefix=None):
    """
    Detect and rewrite tensor expressions into explicit loops.

    This pass analyzes pointwise and reduction tensor expressions and 
    rewrites them into explicit loop-based code, assuming all required 
    arrays have already been allocated.

    Parameters
    ----------
    tree : ast.AST
        The input Python AST containing tensor expressions.
    runtime_vals : dict
        A mapping from variable names to runtime values, used for shape
        analysis.
    loop_index_prefix : str, optional
        Prefix to use for generated loop indices.

    Examples
    --------
    ::

        import ast
        import numpy as np
        from astpass.passes import array_expr_to_loop

        tree = ast.parse("c = a + b")
        rt_vals = {
            'a': np.random.randn(10),
            'b': 8,
            'c': np.empty(10)
        }

        tree = array_expr_to_loop.transform(tree, rt_vals)
        print(ast.unparse(tree))

    Notes
    -----
    This pass assumes that all arrays have been allocated beforehand.
    It only generates loop constructs and does not perform any memory
    allocation. All variables appearing in the input code are assumed
    to be already defined. If the input code inherently requires memory
    allocation for intermediate results, an exception will be raised.
    """
    shape_info = shape_analysis.analyze(tree, runtime_vals)
    return ReductionAndPWExprToLoop(shape_info, loop_index_prefix).visit(tree)



import ast

class HoistShapeAttr(ast.NodeTransformer):
    '''
    Updates `a.shape[0]` to `a_shape_0` and inserts an assignment `a_shape_0 = a.shape[0]`
    before the loop.
    '''
    def __init__(self):
        self.hoisted_shapes = None

    def visit_Subscript(self, node):
        self.generic_visit(node)
        if (
            isinstance(node.value, ast.Attribute)
            and isinstance(node.value.value, ast.Name)
            and node.value.attr == 'shape'
            and isinstance(node.slice, ast.Constant)
        ):
            array_name = node.value.value.id
            dim_index = node.slice.value
            new_name = f"{array_name}_shape_{dim_index}"
            tos = self.hoisted_shapes
            if (array_name, dim_index) not in tos:
                tos.append((array_name, dim_index))            
            return ast.Name(id=new_name, ctx=ast.Load())
        else:
            return node

    def visit_For(self, node):
        # If an outer loop is already being visited by the pass, 
        # treat the inner loops as normal nodes
        if self.hoisted_shapes != None:
            self.generic_visit(node)
            return node

        self.hoisted_shapes = []
        self.generic_visit(node)
        new_assignments = []
        for array_name, dim_index in self.hoisted_shapes:
            new_name = f"{array_name}_shape_{dim_index}"
            shape_access = ast.Subscript(
                value=ast.Attribute(
                    value=ast.Name(id=array_name, ctx=ast.Load()),
                    attr='shape',
                    ctx=ast.Load()
                ),
                slice=ast.Constant(value=dim_index),
                ctx=ast.Load()
            )
            assign = ast.Assign(
                targets=[ast.Name(id=new_name, ctx=ast.Store())],
                value=shape_access
            )
            new_assignments.append(assign)
        self.hoisted_shapes = None
        return new_assignments + [node]
    
def transform(tree):
    '''
    Transforms the AST by hoisting shape attribute accesses.
    '''
    tree = HoistShapeAttr().visit(tree)
    ast.fix_missing_locations(tree)
    return tree